#!/usr/bin/env python

# Copyright 2015 Michael Gugino
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Defines a default resource type'''
import json

class DefaultResource(object):

    def __init__(self):
        '''Initializes a Resource.
           user_objects is a convienence store of user objects and data,
           ideally initialized at the time of calling the app.  This allows us
           to access long-lived DB or Cache sessions, or other long lived
           objects and data.'''

        self.count = 0

    def __call__(self,req,resp,user_objects=None):
        if req.request_method == 'POST':
            self.on_post(req,resp,user_objects)
        elif req.request_method == 'GET':
            self.on_get(req,resp,user_objects)
        else:
            resp.status = '400 Bad Method'
            resp.body = 'Method Not Implemented %s ' % json.dumps(req.url_vars)
            resp.add_header('Content-Type','application/json')

    def on_get(self,req,resp,user_objects=None):
        pass

    def on_post(self,req,resp,user_objects=None):
        pass



#!/usr/bin/env python

# Copyright 2015 Michael Gugino
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Adding routes to your project is easy'''
import pylw.route_find
class Node(object):
    '''Implements basic class for url nodes.  These nodes contain child nodes,
       variable child nodes, and/or functions.  This allows us to walk through
       the requested URL path and get the most relevant result.'''

    def __init__(self,resource=None,name=None,isvar=False):
        self.children = {}
        self.resource = resource
        self.isvar = isvar
        self.name = name
        self.varchild = None

    def __repr__(self): # pragma: no cover
        return self.name

    def get_children(self):
        return self.children

    def get_varchild(self):
        return self.varchild

    def get_resource(self):
        return self.resource

    def get_name(self):
        return self.name

class DefaultRouter(object):
    '''This class implements the DefaultRouter.  DefaultRouter is meant to be
       used directly parse paths into callable resources, and to map requested
       urls from the client to our callable resources we mapped earlier.'''

    def __init__(self):
        self.root_node_dict = {}


    def parse_path(self, uri):
        '''Splits our path into sections using '/' as the separator.
           If the path = '/', then we map '/'.  Otherwise, we trim off the first
           item in the list as it would be blank and nondescriptive.

           The list is returned in reverse so we can pop it.'''
        if uri == '/':
            l = [uri]
        else:
            l = uri.split('/')
            if l[0] == '':
                l = l[1:]
            l.reverse()
        return l

    def add_path(self,uri,resource):
        '''takes in a path string and callable resource, calls parse_path,
           and maps that path to our callable resource.'''

        current_node = None
        root_node = None
        url = self.parse_path(uri)
        while url:
            a = url.pop()

            if root_node is None:
                if not a in self.root_node_dict:
                    root_node = Node(name=a)
                    self.root_node_dict[a] = root_node
                else:
                    root_node = self.root_node_dict[a]

                current_node = root_node

            else: #current_node is not None:
                if '}' not in a:
                    if not a in current_node.children:
                        new_node = Node(name=a)
                        current_node.children[a] = new_node
                    else:
                        new_node = current_node.children[a]
                else:
                    #we found a variable.
                    if not current_node.varchild:
                        new_node = Node(name=a)
                        new_node.isvar = True
                        current_node.varchild = new_node
                    else:
                        new_node = current_node.varchild

                current_node = new_node
        current_node.resource = resource

    def return_path_resource(self,uri,var_dict):
        '''Takes in a string path, and parses it.  It uses the parsed path to
           determine what resource to call.  Raises a 404 exception if a mappped
           resource is not found.

           Also takes in var_dict.  var_dict is an existing dictionary to place
           values of parts of the url that are variables.'''


        current_node = None
        root_node = None
        url = self.parse_path(uri)
        try:
            while url:
                a = url.pop()

                if root_node is None:
                    if not a in self.root_node_dict:
                        #perhaps we should allow a variable on the first node.
                        print "404: no root node: %s" % a
                        raise Exception()
                    else:
                        root_node = self.root_node_dict[a]

                    current_node = root_node

                else:

                    if not a in current_node.children:
                        #we didn't find a defined path, let's try varchild.
                            current_node = current_node.varchild
                    else:
                        current_node = current_node.children[a]

                if current_node.isvar is True:
                    var_dict[current_node.name] = a

            return current_node.resource
        except:
            body = "404: no path found for: %s" % uri
            code = '404 Not Found'
            raise Exception(code, body)

class CRouter(DefaultRouter):

    def parse_path2(self, uri):
        '''Splits our path into sections using '/' as the separator.
           If the path = '/', then we map '/'.  Otherwise, we trim off the first
           item in the list as it would be blank and nondescriptive.

           The list is not returned in reverse'''
        if uri == '/':
            l = [uri]
        else:
            l = uri.split('/')
            if l[0] == '':
                l = l[1:]
        #if the last item is blank due to trailing slash, we must trim it or
        #C will segfault, yay.
        if l[len(l)-1] == '':
            l = l[0:len(l)-2]
        return l

    def return_path_resource(self,uri,var_dict):
        #url = self.parse_path2(uri)
        return pylw.route_find.find_route(uri,self.root_node_dict,var_dict)



#!/usr/bin/env python

# Copyright 2015 Michael Gugino
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''pylw.response.  Contains Response object definition.  Holds body, header,
   and parses/signs cookies.'''

import Cookie

class Response(object):
    '''Response object for sending back to WSGI server.  It holds the headers
       and body'''

    def __init__(self, http_cookies=None, signer=None):
        '''Initialize our object.'''
        self.__header_dict = {}
        self.body = None
        self.status = None
        self.s = signer
        self.__cookies = self.parse_http_cookies(http_cookies)

    def parse_http_cookies(self, req_cookies):
        '''Parse http cookies into a SimpleCookie object'''
        C = Cookie.SimpleCookie()
        if req_cookies:
            C.load(req_cookies)
        return C

    def HTTP404Error(self):
        self.status = '404 Not Found'
        self.body = 'The requested URL was not found.'

    def get_headers(self):
        '''Returns headers in a list of tuples that is usable by WSGI'''
        if not self.body:
            self.body = ''
        self.__header_dict['Content-Length'] = str(len(self.body))

        cookies = []
        if self.__cookies:
            cookies += [("set-cookie", c.OutputString())
            for c in self.__cookies.values()]
        return self.__header_dict.items() + cookies

    def get_cookies(self):
        return self.__cookies.output()

    def get_cookie(self,cookie):
        '''Return the value a cookie.'''
        try:
            return self.__cookies[cookie].value
        except:
            return None

    def get_signed_cookie(self,cookie):
        '''Return the value of a signed cookie.'''
        try:
            return self.s.loads(self.__cookies[cookie].value)
        except:
            return None

    def add_signed_cookie(self,k,v):
        '''Add a signed cookie'''
        v = self.s.dumps(v)
        self.__cookies[k] = v

    def add_cookie(self,k,v):
        '''Add an unsigned cookie'''
        self.__cookies[k] = v

    def add_header(self,k,v):
        '''Add a response header'''
        self.__header_dict[k] = v



from setuptools import setup
from setuptools import setup, find_packages, Extension
from setuptools.command.test import test as TestCommand
import sys, io, os, glob

module1 = Extension('pylw.qs_parse', sources = ['pylw/qs_parse.c'])
module2 = Extension('pylw.route_find', sources = ['pylw/route_find.c'])

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

#Cython Support
try:
    from Cython.Distutils import build_ext
    CYTHON = True
except ImportError:
    CYTHON = False

if CYTHON:
    def list_modules(dirname):
        filenames = glob.glob(os.path.join(dirname, '*.py'))

        module_names = []
        for name in filenames:
            module, ext = os.path.splitext(os.path.basename(name))
            if module != '__init__':
                module_names.append(module)

        return module_names

    ext_modules = [
        Extension('pylw.' + ext, [os.path.join('pylw', ext + '.py')])
        for ext in list_modules(os.path.join('.', 'pylw'))]

    ext_modules.append(module1)
    ext_modules.append(module2)

    cmdclass = {'build_ext': build_ext, 'test': PyTest}

else:
    cmdclass = {'test': PyTest}
    ext_modules = [module1, module2]

setup(name='pylw',
      version='0.1.10',
      description='The Python Lightweight Web Framework',
      long_description=io.open('README.rst', 'r', encoding='utf-8').read(),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Natural Language :: English',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Topic :: Internet :: WWW/HTTP :: WSGI',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
          'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
          'Programming Language :: Python',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: C',
      ],
      url='http://github.com/michaelgugino/pylw',
      author='Michael Gugino',
      author_email='mike@funwithlinux.net',
      tests_require=['pytest'],
      cmdclass=cmdclass,
      license='Apache',
      packages=['pylw'],
      ext_modules = ext_modules,
      zip_safe=False)



'''pylw.request.  Contains Request object definition for parsing environment
   input from WSGI.'''

import Cookie
#import urlparse
import pylw.qs_parse

class Request(object):
    '''An object that holds request values'''

    def __init__(self,env):
        '''Initializes the request object.'''
        self.raw_env = env
        self.request_method = env['REQUEST_METHOD']
        self.posted_body = None
        self.path = env['PATH_INFO']
        self.url_vars = {}

        #uncomment the next line if you want to use python to parse query string.
        #This is probably safer than the C method, and it will handle url encoded
        #strings as well, whereas the c parser will not.
        #self.query_dict = urlparse.parse_qs(env['QUERY_STRING'],keep_blank_values=True)
        if env['QUERY_STRING'] == '':
            self.query_dict = {}
        else:
            self.query_dict = pylw.qs_parse.parse_qs(env['QUERY_STRING'])

        #self.read_post_body()
        if 'HTTP_COOKIE' in env:
            self.cookies = env['HTTP_COOKIE']
        else:
            self.cookies = None

    def read_post_body(self, max_read=4096):
        '''Read the body of the POST request, default max of 4096 bytes'''
        if self.request_method == 'POST':
            stream = self.raw_env['wsgi.input']
            blen = int(self.raw_env['CONTENT_LENGTH'])
            if blen > max_read:
                blen = max_read
            self.posted_body = stream.read(blen)
            stream.close()

    def get_cookies(self):
        '''return self.cookies'''
        return self.cookies



#!/usr/bin/env python

# Copyright 2015 Michael Gugino
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''pylw.app.  Contains Resp,Req,App object definitions.
   pylw.app.App() is the WSGI callable object.'''
import itsdangerous

import routing
import request
import response

class App(object):
    '''This class implements a bare minimum WSGI application'''
    def __init__(self, secret_key=None, config_dict=None, user_objects=None):
        self.router = routing.DefaultRouter()
        self.user_objects = user_objects
        self.hard_coded_path = {}
        self.resp_class = response.Response
        self.req_class = request.Request
        self.s = self.create_secret_signer(secret_key=secret_key)
        if config_dict:
            self.parse_config_dict(config_dict)
        if not secret_key:
            raise Exception('secret key missing')

    def parse_config_dict(self, config_dict):
        '''Override self.__dict__[x] for x in config_dict'''
        for k,v in config_dict.iteritems():
            self.__dict__[k] = config_dict[k]

    def create_secret_signer(self,secret_key=None):
        '''Create our secret signer object'''
        return itsdangerous.Serializer(secret_key)

    def add_hard_coded_path(self,uri,resource):
        '''Add a hard coded path to the application.  You cannot use variables
           here, but the look up time will be much faster than normal.'''

        self.hard_coded_path[uri] = resource

    def __call__(self, env, start_response):
        '''This method is called by the WSGI server.'''


        req = self.req_class(env)
        resp = self.resp_class (
            http_cookies=req.get_cookies(),
            signer=self.s)
        try:
            self.hard_coded_path[req.path](req,resp,user_objects=self.user_objects)
            print self.user_objects
        except:
            try:
                self.router.return_path_resource(
                    req.path,req.url_vars)(req,resp,user_objects=self.user_objects)
            except:
                resp.HTTP404Error()

        start_response(resp.status, resp.get_headers())
        return resp.body
